/*=========================================================================

  Program:   Ionization FRont Interactive Tool (IFRIT)
  Language:  C++


Copyright (c) 2002-2006 Nick Gnedin 
All rights reserved.

This file may be distributed and/or modified under the terms of the
GNU General Public License version 2 as published by the Free Software
Foundation and appearing in the file LICENSE.GPL included in the
packaging of this file.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

=========================================================================*/


#include "idatasubject.h"


#include "ibuffer.h"
#include "icontrolmodule.h"
#include "idata.h"
#include "idataconsumer.h"
#include "idatalimits.h"
#include "idatareader.h"
#include "ierror.h"
#include "ierrorstatus.h"
#include "ieventobserver.h"
#include "ifile.h"
#include "ishellfactory.h"
#include "isystem.h"
#include "iviewmodule.h"

#include <vtkDataSet.h>

//
//  Templates
//
#include "iarraytemplate.h"
#include "ibuffertemplate.h"


#define	IDATASUBJECT_DEFINE_SWAP_BYTES_FUNCTIONS(_type_) \
void iDataSubject::SwapBytes(_type_ &p) \
{ \
	iDataSubject_Private::SwapBytes(p); \
} \
void iDataSubject::SwapBytesRange(_type_ *p, long count) \
{ \
	iDataSubject_Private::SwapBytesRange(p,count); \
}

#define	IDATASUBJECT_DEFINE_READ_BLOCK_FUNCTIONS(_type_) \
bool iDataSubject::ReadBlock(iFile& F, _type_ *p, long len, float updateStart, float updateDuration) \
{ \
	return iDataSubject_Private::ReadBlockToArray(F,p,len,updateStart,updateDuration,mIsBigEndian!=iSystem::IsBigEndianMachine(),mObserver); \
} \
iDataSubject::ReadingBuffer iDataSubject::Buffer(_type_ &d) const \
{ \
	return ReadingBuffer(&d,1UL,iDataSubject_Private::TypeIndex(&d),sizeof(_type_)); \
} \
iDataSubject::ReadingBuffer iDataSubject::Buffer(_type_ *d, long l) const \
{ \
	return ReadingBuffer(d,l,iDataSubject_Private::TypeIndex(d),sizeof(_type_)); \
}


namespace iDataSubject_Private
{
	//
	//  Byte swap helpers
	//
	inline void Swap4Bytes(char* data)
	{ 
		char one_byte; 
		one_byte = data[0]; data[0] = data[3]; data[3] = one_byte;
		one_byte = data[1]; data[1] = data[2]; data[2] = one_byte; 
	}


	inline void Swap8Bytes(char* data)
	{ 
		char one_byte;
		one_byte = data[0]; data[0] = data[7]; data[7] = one_byte;
		one_byte = data[1]; data[1] = data[6]; data[6] = one_byte;
		one_byte = data[2]; data[2] = data[5]; data[5] = one_byte;
		one_byte = data[3]; data[3] = data[4]; data[4] = one_byte; 
	}

	
	void Swap4BytesRange(char *data, long count)
	{
		char *pos;
		long i;

		pos = data;
		for(i = 0; i<count; i++)
		{
			Swap4Bytes(pos);
			pos += 4;
		}
	}


	void Swap8BytesRange(char *data, long count)
	{
		char *pos;
		long i;

		pos = data;
		for(i = 0; i<count; i++)
		{
			Swap8Bytes(pos);
			pos += 8;
		}
	}


	template<class T>
	inline void SwapBytes(T &p)
	{
		switch(sizeof(T))
		{
		case 4:
			{
				Swap4Bytes((char *)&p);
				break;
			}
		case 8:
			{
				Swap8Bytes((char *)&p);
				break;
			}
		default:
			{
				IERROR_FINISH("IFrIT was not ported to that platform.");
			}
		}
	}


	template<class T>
	inline void SwapBytesRange(T *p, long count)
	{
		switch(sizeof(T))
		{
		case 4:
			{
				Swap4BytesRange((char *)p,count);
				break;
			}
		case 8:
			{
				Swap8BytesRange((char *)p,count);
				break;
			}
		default:
			{
				IERROR_FINISH("IFrIT was not ported to that platform.");
			}
		}
	}


	//
	//  Check validity of numbers
	//
	bool IsNumberValid(int v){ return true; }
	bool IsNumberValid(long v){ return true; }
	bool IsNumberValid(float v){ return (-iMath::_FloatMax<=v && v<=iMath::_FloatMax); }
	bool IsNumberValid(double v){ return (-iMath::_DoubleMax<=v && v<=iMath::_DoubleMax); }


	//
	//  Type indices
	//
	const int _Int = 1;
	const int _Long = 2;
	const int _Float = 3;
	const int _Double = 4;

	int TypeIndex(int *){ return _Int; }
	int TypeIndex(long *){ return _Long; }
	int TypeIndex(float *){ return _Float; }
	int TypeIndex(double *){ return _Double; }


	//
	//  Read blocks of data from a file
	//
	template<class T>
	bool ReadBlockToArray(iFile& F, T *p, long len, float updateStart, float updateDuration, bool swapbytes, iProgressEventObserver *observer)
	{
		const long nline = 4096;
		long nread, l, nstep = (len+nline-1)/nline;
		float ustep = updateDuration/nstep;

		for(l=0; l<nstep; l++) 
		{
			if(observer->IsAborted()) return true;
			if(l < nstep-1) nread = nline; else nread  = len - nline*(nstep-1);
			if(!F.ReadBlock((char *)(p+l*nline),nread*sizeof(T))) return false;
			observer->SetProgress(updateStart+ustep*l);
		}

		if(swapbytes) 
		{
			iDataSubject_Private::SwapBytesRange(p,len);
		}

		bool ok = true;
		for(l=0; ok && l<len; l++)
		{
			ok = iDataSubject_Private::IsNumberValid(p[l]);
		}

		return ok;
	}


	bool ReadBlockToBuffer(iFile& F, const iDataSubject::ReadingBuffer &b, float updateStart, float updateDuration, bool swapbytes, iProgressEventObserver *observer)
	{
		switch(b.Type)
		{
		case _Int:
			{
				int *d = (int *)b.Data;
				return iDataSubject_Private::ReadBlockToArray(F,d,b.Length,updateStart,updateDuration,swapbytes,observer);
			}
		case _Long:
			{
				long *d = (long *)b.Data;
				return iDataSubject_Private::ReadBlockToArray(F,d,b.Length,updateStart,updateDuration,swapbytes,observer);
			}
		case _Float:
			{
				float *d = (float *)b.Data;
				return iDataSubject_Private::ReadBlockToArray(F,d,b.Length,updateStart,updateDuration,swapbytes,observer);
			}
		case _Double:
			{
				double *d = (double *)b.Data;
				return iDataSubject_Private::ReadBlockToArray(F,d,b.Length,updateStart,updateDuration,swapbytes,observer);
			}
		default:
			{
				return false;
			}
		}
	}


	template<class T>
	bool AnalyzeFirstRecordForType(iFile &F, unsigned int len, bool &isBig)
	{
		T lrec1, lrec2;

		int m = F.SetMarker();
		if(!F.ReadBlock((char *)&lrec1,sizeof(T)) || !F.SkipBlock(len) || !F.ReadBlock((char *)&lrec2,sizeof(T)))
		{
			F.ReturnToMarker(m,true);
			return false; // unable even to read
		}
		F.ReturnToMarker(m,true);

		//
		//  Is the header/footer length ok?
		//
		if(lrec1 == lrec2) // succeeded
		{
			//
			// auto-detect data endiness
			//
			if(lrec1 != len)
			{
				SwapBytes(lrec1);
				if(lrec1 != len) return false; else isBig = !iSystem::IsBigEndianMachine();
			}
			else isBig = iSystem::IsBigEndianMachine();
			return true;
		}
		else return false;
	}
};


using namespace iDataSubject_Private;
using namespace iParameter;


IOBJECT_DEFINE_TYPE(iDataSubject,Data,-d,iObjectType::_Helper);


IOBJECT_DEFINE_KEY(iDataSubject,DataPresent,dp,Bool,1);
IOBJECT_DEFINE_KEY(iDataSubject,FixedLimits,fl,Bool,1);
IOBJECT_DEFINE_KEY(iDataSubject,FileName,fn,String,1);
IOBJECT_DEFINE_KEY(iDataSubject,FixedStretch,fs,Bool,0);
IOBJECT_DEFINE_KEY(iDataSubject,LowerLimit,lo,Float,0);
IOBJECT_DEFINE_KEY(iDataSubject,Max,max,Float,0);
IOBJECT_DEFINE_KEY(iDataSubject,Min,min,Float,0);
IOBJECT_DEFINE_KEY(iDataSubject,NumVars,n,Int,1);
IOBJECT_DEFINE_KEY(iDataSubject,Name,na,String,0);
IOBJECT_DEFINE_KEY(iDataSubject,Stretch,s,Int,0);
IOBJECT_DEFINE_KEY(iDataSubject,UpperLimit,up,Float,0);

IOBJECT_DEFINE_KEY(iDataSubject,ResizeLimits,-rl,Int,1); // hidden key
IOBJECT_DEFINE_KEY(iDataSubject,RecordLength,-l,Int,1);  // hidden key


bool iDataSubject::IsTypeMatch(const iString &prefix)
{
	iString pre = prefix.Section("-",0,0);
	if(pre == "d") return true; else return iDataSubject::Type().IsMatch(pre);
}


iString iDataSubject::GetDataSubjectName(const iString &prefix)
{
	static iString none;
	//
	//  Set the subject name
	//
	int i1 = prefix.Find("-");
	if(i1 < 0) return none;
	int i2 = prefix.Find(iObjectKey::PrefixSeparator());
	if(i2 < 0) i2 = prefix.Length();
	return prefix.Part(i1+1,i2-i1-1);
}


iDataSubject::iDataSubject(iDataReader *r, const iDataType &type) : iViewModuleComponent(r->GetViewModule()), mDataType(type)
{
	mRecordLength = 4; 
	mFortranHeaderFooterLength = -1;

	mReader = r;
	mLimits = 0;

	if(mDataType == iData::None())
	{
		IERROR_FINISH("DataSubject cannot have Null DataType.");
	}

	mInternalData = mExternalData = 0;  // must be set by a child

	// mFileRoot = mFileSuffix = mLastFileName = "";
	mUpdated = true;
	mIsBigEndian = true;
	mTwoCopies = false;
	mOwnsData = type.IsPrimary();
	mRecord = -1;
	mShift[0] = mShift[1] = mShift[2] = 0.0f;

	mPeriodic[0] = mPeriodic[1] = mPeriodic[2] = false;
	mBoundaryConditions = _BoundaryConditionsWall;

	mAdjustableLimits = true;

	//
	//  Event observers are driven directly (otherwise we would have to create at least 16 new events, which is
	//  cumbersome), so adding with AddObserver is not necessarily
	//
	mObserver = dynamic_cast<iProgressEventObserver *>(iShellFactory::CreateEventObserver("Progress",mReader->GetViewModule())); IERROR_ASSERT(mObserver);
}


iDataSubject::~iDataSubject()
{
	if(mLimits != 0) delete mLimits;

	mObserver->Delete();
	if(mOwnsData)
	{
		if(mInternalData != 0) mInternalData->Delete();
		if(mExternalData != 0) mExternalData->Delete();
	}
}


const iObjectType& iDataSubject::RealType()
{
	return iDataSubject::Type();
}


bool iDataSubject::IsUsingData(const iDataType &type) const
{
	return mDataType == type;
}


vtkDataSet* iDataSubject::GetData() const
{
	return mExternalData;
}


void iDataSubject::SetBoundaryConditions(int s)
{
	if(s>=0 && s<__NumBoundaryConditions)
	{
		mBoundaryConditions = s;
	}
}


void iDataSubject::SetDirectionPeriodic(int d, bool s)
{
	if(d>=0 && d<3)
	{
		mPeriodic[d] = s;
	}
}


iDataLimits* iDataSubject::GetLimits() const
{
	if(mLimits == 0)
	{
		mLimits = this->CreateLimits(); IERROR_ASSERT(mLimits);
	}
	
	return mLimits;
}


void iDataSubject::ResetLimits()
{
	mLimits = 0;
}


void iDataSubject::ReadFile(const iString &fname)
{
	this->GetErrorStatus()->Clear();
	mOverflow = false;
	mFortranHeaderFooterLength = -1; // must be set by a child class

	if(!mOwnsData)
	{
		this->GetErrorStatus()->Set("Cannot load a data in a subject that does not own its data.");
		return;
	}

	iString root, suffix;
	int record = this->DissectFileName(fname,root,suffix);

	this->GetLimits()->BlockNotifications(true);
	this->ReadFileBody(suffix,fname);
	this->GetLimits()->BlockNotifications(false);

	if(this->GetErrorStatus()->NoError())
	{
		mLastFileName = fname;
		mFileSuffix = suffix;
		mFileRoot = root;
		mUpdated = false;
		mRecord = record;
		mShift[0] = mShift[1] = mShift[2] = 0.0f;
	}
}


void iDataSubject::Finalize()
{
	this->ClearCache();

	this->GetLimits()->BlockNotifications(true);
	this->FinalizeBody();
	this->GetLimits()->BlockNotifications(false);

	if(!mUpdated && mExternalData!=0 && mInternalData!=0 && mInternalData->GetNumberOfPoints()>0)
	{
		mUpdated = true;

		if(mTwoCopies)
		{
			mExternalData->DeepCopy(mInternalData);
		}
		else
		{
			mExternalData->ShallowCopy(mInternalData);
		}

		this->Polish();

		if(this->GetErrorStatus()->NoError() && mOverflow)
		{
			this->GetErrorStatus()->Set("The data are read correctly, but some values in the file were outside of range ("+iString::FromNumber(-iMath::_LargeFloat)+","+iString::FromNumber(iMath::_LargeFloat)+").\n These values were clamped to be within the required range.",-1);
		}
	}

	this->NotifyDependencies();
}


void iDataSubject::NotifyDependencies()
{
	//
	//  Default implementation, but can be changed in a child
	//
	this->GetLimits()->NotifyDependencies(-1);
}


bool iDataSubject::IsAborted() const
{
	return mObserver->IsAborted();
}


bool iDataSubject::IsThereData() const
{
	vtkDataSet *data = this->GetData();
	return (data!=0 && data->GetNumberOfPoints() > 0);
}


void iDataSubject::EraseData()
{
	if(mOwnsData)
	{
		if(mInternalData != 0) mInternalData->Initialize();
		if(mExternalData != 0) mExternalData->Initialize();
	}
	mRecord = -1;
	mShift[0] = mShift[1] = mShift[2] = 0.0f;

	this->GetLimits()->AssignVars(); // if no data, use all vars
	this->ClearCache();
}


void iDataSubject::DetachData()
{
	if(mTwoCopies) return; // if we keep two copies, the data are already detached

	if(mExternalData != 0) mExternalData->Initialize();
}


void iDataSubject::Reset()
{
	if(mExternalData != 0) mExternalData->Modified();
	this->GetViewModule()->NotifyDataConsumers(this->Request());
}


void iDataSubject::ShiftData(double dr[3])
{
	int i;

	mObserver->Started(iProgressEventObserver::_Shifting);
	mObserver->SetProgress(0.0);

	for(i=0; i<3; i++) if(fabs(dr[i]-mShift[i]) > 1.0e-100)
	{
		this->ShiftDataBody(i,dr[i]-mShift[i]);
		mShift[i] = dr[i];
	}

	this->Polish();

	if(mExternalData!=0 && mInternalData!=0)
	{
		if(mTwoCopies)
		{
			mExternalData->DeepCopy(mInternalData);
		}
		else
		{
			mExternalData->ShallowCopy(mInternalData);
		}
	}

	mObserver->Finished();
	this->ClearCache();
}


void iDataSubject::SetAdjustableLimits(bool s)
{
	mAdjustableLimits = s;
}


float iDataSubject::GetMemorySize() const
{
	if(mOwnsData)
	{
		float s = 0.0f;
		if(mInternalData != 0) s += mInternalData->GetActualMemorySize();
		if(mTwoCopies && mExternalData!=0) s += mExternalData->GetActualMemorySize();
		return s;
	}
	else return 0.0f;
}


//
//  Data reading helpers
//
bool iDataSubject::SkipFortranRecord(iFile& F, long len)
{
	long lrec1, lrec2;

	if(!this->ReadFortranHeaderFooter(F,lrec1)) return false;
	if(!F.SkipBlock(len)) return false;
	if(!this->ReadFortranHeaderFooter(F,lrec2)) return false;
	if(lrec1!=lrec2 || lrec1!=len) return false;

	return true;
}


bool iDataSubject::ReadFortranRecord(iFile& F, const ReadingBuffer &b, float updateStart, float updateDuration, bool autoswap)
{
	long lrec1, lrec2;

	if(!this->ReadFortranHeaderFooter(F,lrec1)) return false;

	if(!ReadBlockToBuffer(F,b,updateStart,updateDuration,autoswap && mIsBigEndian!=iSystem::IsBigEndianMachine(),mObserver)) return false;
	if(mObserver->IsAborted()) return true;

	if(!this->ReadFortranHeaderFooter(F,lrec2)) return false;
	if(lrec1!=lrec2 || lrec1!=b.Length*b.Size) return false;

	return true;
}


bool iDataSubject::ReadFortranRecord(iFile& F, const ReadingBuffer &b1, const ReadingBuffer &b2, float updateStart, float updateDuration, bool autoswap)
{
	long lrec1, lrec2;
	long len = b1.Length*b1.Size + b2.Length*b2.Size;
	float f1 = float(b1.Length*b1.Size)/len;
	float f2 = float(b2.Length*b2.Size)/len;

	if(!this->ReadFortranHeaderFooter(F,lrec1)) return false;

	if(!ReadBlockToBuffer(F,b1,updateStart,f1*updateDuration,autoswap && mIsBigEndian!=iSystem::IsBigEndianMachine(),mObserver)) return false;
	if(mObserver->IsAborted()) return true;
	if(!ReadBlockToBuffer(F,b2,updateStart+f1*updateDuration,f2*updateDuration,autoswap && mIsBigEndian!=iSystem::IsBigEndianMachine(),mObserver)) return false;
	if(mObserver->IsAborted()) return true;

	if(!this->ReadFortranHeaderFooter(F,lrec2)) return false;
	if(lrec1!=lrec2 || lrec1!=len) return false;

	return true;
}


bool iDataSubject::ReadFortranRecord(iFile& F, const ReadingBuffer &b1, const ReadingBuffer &b2, const ReadingBuffer &b3, float updateStart, float updateDuration, bool autoswap)
{
	long lrec1, lrec2;
	long len = b1.Length*b1.Size + b2.Length*b2.Size + b3.Length*b3.Size;
	float f1 = float(b1.Length*b1.Size)/len;
	float f2 = float(b2.Length*b2.Size)/len;
	float f3 = float(b3.Length*b3.Size)/len;

	if(!this->ReadFortranHeaderFooter(F,lrec1)) return false;

	if(!ReadBlockToBuffer(F,b1,updateStart,f1*updateDuration,autoswap && mIsBigEndian!=iSystem::IsBigEndianMachine(),mObserver)) return false;
	if(mObserver->IsAborted()) return true;
	if(!ReadBlockToBuffer(F,b2,updateStart+f1*updateDuration,f2*updateDuration,autoswap && mIsBigEndian!=iSystem::IsBigEndianMachine(),mObserver)) return false;
	if(mObserver->IsAborted()) return true;
	if(!ReadBlockToBuffer(F,b3,updateStart+(f1+f2)*updateDuration,f3*updateDuration,autoswap && mIsBigEndian!=iSystem::IsBigEndianMachine(),mObserver)) return false;
	if(mObserver->IsAborted()) return true;

	if(!this->ReadFortranHeaderFooter(F,lrec2)) return false;
	if(lrec1!=lrec2 || lrec1!=len) return false;

	return true;
}

	
bool iDataSubject::ReadFortranRecord(iFile& F, int nbufs, const ReadingBuffer *b, float updateStart, float updateDuration, bool autoswap) // read any number of buffers
{
	long lrec1, lrec2;

	if(b==0 || nbufs<=0 || !this->ReadFortranHeaderFooter(F,lrec1)) return false;

	int i;
	long len = 0L;
	for(i=0; i<nbufs; i++)
	{
		len += b[i].Length*b[i].Size;
		if(!ReadBlockToBuffer(F,b[i],updateStart+i*updateDuration,updateDuration,autoswap && mIsBigEndian!=iSystem::IsBigEndianMachine(),mObserver)) return false;
		if(mObserver->IsAborted()) return true;
	}

	if(!this->ReadFortranHeaderFooter(F,lrec2)) return false;
	if(lrec1!=lrec2 || lrec1!=len) return false;

	return true;
}


bool iDataSubject::ReadFortranHeaderFooter(iFile& F, long &lrec)
{
	bool ret = false;

	if(mFortranHeaderFooterLength == sizeof(long))
	{
		ret = F.ReadBlock((char *)&lrec,sizeof(long));
		if(ret && mIsBigEndian!=iSystem::IsBigEndianMachine()) this->SwapBytes(lrec);
	}
	else if(mFortranHeaderFooterLength == sizeof(int))
	{
		int ltmp;
		ret = F.ReadBlock((char *)&ltmp,sizeof(int));
		if(ret)
		{
			if(mIsBigEndian != iSystem::IsBigEndianMachine()) this->SwapBytes(ltmp);
			lrec = (long)ltmp;
		}
	}
	else
	{
		IERROR_REPORT("Invalid Fortran header/footer length.");
	}

	return ret;
}


void iDataSubject::PackCompleteState(iString &s) const
{
	iDataLimits *limits = this->GetLimits();
	//
	//  Save the currently loaded Vars
	//
	limits->BackupVars();
	limits->BlockNotifications(true);
	limits->AssignVars();
	limits->BlockNotifications(false);
	//
	//  Pack state without caching
	//
	s.Clear();
	this->PackValue(s,KeyResizeLimits(),limits->GetSize());
	this->PackStateBody(s);
	//
	//  Restore the currently loaded Vars
	//
	limits->RestoreVars();
}


void iDataSubject::UnPackCompleteState(const iString &s)
{
	iDataLimits *limits = this->GetLimits();
	//
	//  Save the currently loaded Vars
	//
	limits->BackupVars();
	limits->BlockNotifications(true);
	limits->AssignVars();
	limits->BlockNotifications(false);
	//
	//  Unpack state and clear the cache
	//
	this->UnPackStateBody(s);
	this->ClearCache();
	//
	//  Restore the currently loaded Vars
	//
	limits->RestoreVars();
}


void iDataSubject::PackStateBody(iString &s) const
{
	static iBuffer<int> itmp;
	static iBuffer<bool> btmp;
	static iBuffer<float> ftmp;
	static iBuffer<iString> stmp;
	int i, n;

	iDataLimits *limits = this->GetLimits();

	this->PackValue(s,KeyFixedLimits(),!this->GetAdjustableLimits());
	this->PackValue(s,KeyDataPresent(),this->IsThereData());
	this->PackValue(s,KeyFileName(),this->GetLastFileName());

	n = limits->GetNumVars();
	this->PackValue(s,KeyNumVars(),n);

	if(n > 0)
	{
		for(i=0; i<n; i++) itmp[i] = limits->GetStretch(i); 
		this->PackValue(s,KeyStretch(),itmp,n);

		for(i=0; i<n; i++) ftmp[i] = limits->GetMin(i);
		this->PackValue(s,KeyMin(),ftmp,n);

		for(i=0; i<n; i++) ftmp[i] = limits->GetMax(i);
		this->PackValue(s,KeyMax(),ftmp,n);

		for(i=0; i<n; i++) ftmp[i] = limits->GetLowerLimit(i);
		this->PackValue(s,KeyLowerLimit(),ftmp,n);

		for(i=0; i<n; i++) ftmp[i] = limits->GetUpperLimit(i);
		this->PackValue(s,KeyUpperLimit(),ftmp,n);

		for(i=0; i<n; i++) btmp[i] = limits->GetFixedStretch(i);
		this->PackValue(s,KeyFixedStretch(),btmp,n);

		for(i=0; i<n; i++) stmp[i] = limits->GetName(i);
		this->PackValue(s,KeyName(),stmp,n);
	}

	//
	//  Hidden key
	//
	this->PackValue(s,KeyRecordLength(),mRecordLength);

	this->DataSubjectPackStateBody(s);
}


void iDataSubject::UnPackStateBody(const iString &s)
{
	static iBuffer<int> itmp;
	static iBuffer<float> ftmp;
	static iBuffer<iString> stmp;
	int i, n; bool b;
	
	iDataLimits *limits = this->GetLimits();
	limits->BlockNotifications(true);

	if(this->UnPackValue(s,KeyResizeLimits(),n))
	{
		limits->Resize(n);
		if(!this->IsThereData()) limits->AssignVars();
		this->ClearCache();
	}
	if(this->UnPackValue(s,KeyFixedLimits(),b)) this->SetAdjustableLimits(!b);

	n = limits->GetNumVars();
	if(this->UnPackValue(s,KeyNumVars(),n))
	{
		this->ClearCache();
		n = limits->GetNumVars();
	}

	if(n > 0)
	{
		for(i=0; i<n; i++) itmp[i] = limits->GetStretch(i);
		if(this->UnPackValue(s,KeyStretch(),itmp,n))
		{
			for(i=0; i<n; i++) limits->SetStretch(i,itmp[i]);
			this->ClearCache();
		}

		for(i=0; i<n; i++) ftmp[i] = limits->GetMin(i);
		if(this->UnPackValue(s,KeyMin(),ftmp,n))
		{
			for(i=0; i<n; i++) limits->SetMin(i,ftmp[i]);
			this->ClearCache();
		}

		for(i=0; i<n; i++) ftmp[i] = limits->GetMax(i);
		if(this->UnPackValue(s,KeyMax(),ftmp,n))
		{
			for(i=0; i<n; i++) limits->SetMax(i,ftmp[i]);
			this->ClearCache();
		}

		for(i=0; i<n; i++) ftmp[i] = limits->GetLowerLimit(i);
		if(this->UnPackValue(s,KeyLowerLimit(),ftmp,n))
		{
			for(i=0; i<n; i++) limits->SetLowerLimit(i,ftmp[i]);
			this->ClearCache();
		}

		for(i=0; i<n; i++) ftmp[i] = limits->GetUpperLimit(i);
		if(this->UnPackValue(s,KeyUpperLimit(),ftmp,n))
		{
			for(i=0; i<n; i++) limits->SetUpperLimit(i,ftmp[i]);
			this->ClearCache();
		}

		for(i=0; i<n; i++) stmp[i] = limits->GetName(i);
		if(this->UnPackValue(s,KeyName(),stmp,n))
		{
			for(i=0; i<n; i++) limits->SetName(i,stmp[i]);
			this->ClearCache();
		}
	}

	//
	//  Hidden key
	//
	iObject::ReportMissingKeys(false);
	if(this->UnPackValue(s,KeyRecordLength(),i))
	{
		mRecordLength = i;
		this->ClearCache();
	}
	iObject::ReportMissingKeys(true);

	this->DataSubjectUnPackStateBody(s);

	limits->BlockNotifications(false);
	limits->NotifyDependencies(-1);
}


bool iDataSubject::DetectFortranFileStructure(iFile &F, long len)
{
	//
	//  Try int-sized records
	//
	if(AnalyzeFirstRecordForType<int>(F,len,mIsBigEndian))
	{
		mFortranHeaderFooterLength = sizeof(int);
		return true;
	}
	//
	//  Try long-sized records
	//
	if(AnalyzeFirstRecordForType<long>(F,len,mIsBigEndian))
	{
		mFortranHeaderFooterLength = sizeof(long);
		return true;
	}
	//
	//  Failed to figure it out
	//
	return false;
}


const iDataSyncRequest& iDataSubject::Request(int var) const
{
	return iDataConsumer::Request(this->GetDataType(),var);
}


iString iDataSubject::GetRecordAsString(int rec) const
{
	const iString format = "%0" + iString::FromNumber(mRecordLength) + "d";
	return iString::FromNumber(rec,format.ToCharPointer());
}


int iDataSubject::DissectFileName(const iString &fname, iString &root, iString &suffix) const
{
	bool ok;
	int rec;

	suffix = fname.Section(".",-1);
	root = fname.Part(0,fname.Length()-suffix.Length()-1);
	iString record = root.Part(-mRecordLength);
	rec = record.ToInt(ok);
	if(!ok || record!=this->GetRecordAsString(rec) || rec<0) 
	{
		return -1;
	}
	else
	{
		root = root.Part(0,root.Length()-mRecordLength);
		return rec;
	}
}


const iString iDataSubject::GetFileName(int rec) const
{
	if(mFileSuffix.IsEmpty())
	{
		return mFileRoot + this->GetRecordAsString(rec);
	}
	else
	{
		return mFileRoot + this->GetRecordAsString(rec) + "." + mFileSuffix;
	}
}


bool iDataSubject::IsSeriesFileName(const iString &fname) const
{
	iString root, suffix;
	int	rec = this->DissectFileName(fname,root,suffix);
	return (rec>=0 && root==mFileRoot && suffix==mFileSuffix);
}


//
//  Overloads
//
IDATASUBJECT_DEFINE_SWAP_BYTES_FUNCTIONS(int);
IDATASUBJECT_DEFINE_SWAP_BYTES_FUNCTIONS(long);
IDATASUBJECT_DEFINE_SWAP_BYTES_FUNCTIONS(float);
IDATASUBJECT_DEFINE_SWAP_BYTES_FUNCTIONS(double);
//IDATASUBJECT_DEFINE_SWAP_BYTES_FUNCTIONS(unsigned int);
//IDATASUBJECT_DEFINE_SWAP_BYTES_FUNCTIONS(unsigned long);

IDATASUBJECT_DEFINE_READ_BLOCK_FUNCTIONS(int);
IDATASUBJECT_DEFINE_READ_BLOCK_FUNCTIONS(long);
IDATASUBJECT_DEFINE_READ_BLOCK_FUNCTIONS(float);
IDATASUBJECT_DEFINE_READ_BLOCK_FUNCTIONS(double);
//IDATASUBJECT_DEFINE_READ_BLOCK_FUNCTIONS(unsigned int);
//IDATASUBJECT_DEFINE_READ_BLOCK_FUNCTIONS(unsigned long);


//
//  Secondary subject (for data subjects that do not own their data)
//
iSecondaryDataSubject::iSecondaryDataSubject(iPrimaryDataSubject *parent, const iDataType &type) : iDataSubject(parent->GetReader(),type)
{
	int i;

	mParent = parent; IERROR_ASSERT(parent);
	mOwnsData = false;

	mBoundaryConditions = parent->GetBoundaryConditions();
	for(i=0; i<3; i++) mPeriodic[i] = parent->IsDirectionPeriodic(i);
	mAdjustableLimits = parent->GetAdjustableLimits();

	mParent->RegisterDependentSubject(this);
}


iSecondaryDataSubject::~iSecondaryDataSubject()
{
	this->ResetLimits();  // a copy, not an independent object
	if(mParent != 0) mParent->UnRegisterDependentSubject(this);
}


void iSecondaryDataSubject::DataSubjectPackStateBody(iString &) const
{
}

void iSecondaryDataSubject::DataSubjectUnPackStateBody(const iString &)
{
}


void iSecondaryDataSubject::ReadFileBody(const iString &, const iString &)
{
}


void iSecondaryDataSubject::ShiftDataBody(int, double)
{
}


void iSecondaryDataSubject::FinalizeBody()
{
}


iProbeFilter* iSecondaryDataSubject::CreateProbeFilter(iViewSubject *vo) const
{
	if(mParent != 0) return mParent->CreateProbeFilter(vo); else return 0;
}


iHistogramMaker* iSecondaryDataSubject::CreateHistogramMaker() const
{
	if(mParent != 0) return mParent->CreateHistogramMaker(); else return 0;
}


//
//  Primary subject (for data subjects that control secondary subject)
//
iPrimaryDataSubject::iPrimaryDataSubject(iDataReader *r, const iDataType &type) : iDataSubject(r,type), mDependentSubjects(0)
{
}


iPrimaryDataSubject::~iPrimaryDataSubject()
{
	int i;

	for(i=0; i<mDependentSubjects.Size(); i++) mDependentSubjects[i]->mParent = 0;
}


void iPrimaryDataSubject::NotifyDependencies()
{
	int i;

	this->iDataSubject::NotifyDependencies();
	for(i=0; i<mDependentSubjects.Size(); i++) mDependentSubjects[i]->NotifyDependencies();
}


void iPrimaryDataSubject::RegisterDependentSubject(iSecondaryDataSubject *subject)
{
	if(subject != 0) mDependentSubjects.AddUnique(subject);
}


void iPrimaryDataSubject::UnRegisterDependentSubject(iSecondaryDataSubject *subject)
{
	if(subject != 0) mDependentSubjects.Remove(subject);
}

