/*=========================================================================

  Program:   Ionization FRont Interactive Tool (IFRIT)
  Language:  C++


Copyright (c) 2002-2006 Nick Gnedin 
All rights reserved.

This file may be distributed and/or modified under the terms of the
GNU General Public License version 2 as published by the Free Software
Foundation and appearing in the file LICENSE.GPL included in the
packaging of this file.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

=========================================================================*/


#ifndef ISHELL_H
#define ISHELL_H

//
//  A base class for all GUI shells around the VTK-based core
//

#include "iarray.h"
#include "istring.h"

class iControlModule;
class iFile;
class iObjectKey;
class iOutputWindow;

class vtkTimerLog;


class iShell
{

public:

	struct Option
	{
		iString Name;
		iString Help;
		bool NeedsValue;
		iString Value;
		Option() { NeedsValue = false; }
	};

	//
	//  A helper function that creates, runs, and deletes a shell in one call
	//
	static void RunApplication(const iString &type, int argc, char **argv);

	//
	//  Create a shell with this function
	//
	static iShell* New(const iString &type, int argc, char **argv);

	//
	//  This function does the shell's work after all the windows are created. 
	//
	void Run();

	//
	//  Exit from the current shell 
	//
	virtual void Exit() = 0;

	//
	//  Destroy the shell with this function
	//
	void Delete();

	//
	//  Access to components
	//
	inline const iString& Type() const { return mType; }
	inline iControlModule* GetControlModule() const { return mControlModule; }
	inline iOutputWindow* GetOutputWindow() const { return mOutputWindow; }
	inline bool IsRunning() const { return mRunning; }

	const iString& GetEnvironment(int key) const;
	inline const iString& GetStateFileName() const { return mStateFileName; }

	virtual bool LoadShellStateFromFile(iFile &F) = 0;
	virtual bool SaveShellStateToFile(iFile &F) const  = 0;

	void OnInitAtom();

	//
	//  Directory naming convention for portability
	//
	const iString& GetDirSeparator() const;
	const iString& GetCurrentDir() const;

#ifdef I_DEBUG
	virtual void OutputDebugMessage(const iString &message) const = 0;
#endif

protected:

	iShell(int argc, char **argv);
	virtual ~iShell();  

	//
	//  This functions parses command-line options.
	//
	void ParseCommandLineOptions(int argc, char **argv);  
	void AddCommandLineOption(const iString &root, bool hasValue, const iString &help);
	bool AnalyseOneCommandLineOption(const struct Option &o);
	
	//
	//  Because of cyclic dependence of different objects on each other, the constructor
	//  and destructor functions of a child shell should be empty, and all constructor/destructor 
	//  work must be done in overwrites of the following two functions:
	//
	virtual void ConstructorBody() = 0;  
	virtual void DestructorBody() = 0;  

	//
	// These functions must be overwritten in a child class
	//
	virtual void Start() = 0;
	virtual void RunBody() = 0; // body of Run()
	virtual void PrepareForConstruction() = 0;
	virtual void DefineSpecificCommandLineOptions() = 0;
	virtual void LoadStateFile(const iString &fname) = 0;
	virtual void OutputStartupMessage(const iString &message) = 0;
	virtual bool AnalyseOneExtendedCommandLineOption(const struct Option &o) = 0;

	//
	//  This function can be used to animate the initialization progress.
	//  Does nothing by default
	//
	virtual void OnInitAtomBody(bool timed);

	iControlModule *mControlModule;
	iOutputWindow *mOutputWindow;

	bool mRunning;
	iString mType;
	static const iString mOptionPrefix;
	iArray<Option> mOptions;
	iString mStateFileName;
	int mNumProcs, mFirstOption;

	vtkTimerLog *mInitTimer;
	
	//
	//  Initialization
	//
	struct InitStep
	{
		int Current;
		int Total;
	};

	int mCurInitStep;
	InitStep mInitSteps[4];

	//
	//  Environment
	//
	iString mEnvBaseDir;
	iString mEnvDataDir;
	iString mEnvImageDir;
	iString mEnvScriptDir;
	iString mEnvPaletteDir;

private:

#ifdef I_DEVEL
	void InitHelper();
#endif
};

#endif  // ISHELL_H
