/*=========================================================================

  Program:   Ionization FRont Interactive Tool (IFRIT)
  Language:  C++


Copyright (c) 2002-2006 Nick Gnedin 
All rights reserved.

This file may be distributed and/or modified under the terms of the
GNU General Public License version 2 as published by the Free Software
Foundation and appearing in the file LICENSE.GPL included in the
packaging of this file.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

=========================================================================*/


#include "iviewsubjectparallelpipeline.h"


#include "ierror.h"
#include "iparallel.h"
#include "iparallelmanager.h"
#include "iviewmodule.h"
#include "iviewsubject.h"
#include "iviewsubjectpipelinedatamanager.h"

#include <vtkDataSet.h>
#include <vtkPolyData.h>

//
//  Templates
//
#include "iarraytemplate.h"


iViewSubjectParallelPipeline* iViewSubjectParallelPipeline::New(iViewSubject *vo, int numInputs, int id)
{
	return new iViewSubjectParallelPipeline(vo,numInputs,id);
}


iViewSubjectParallelPipeline::iViewSubjectParallelPipeline(iViewSubject *vo, int numInputs, int id) : iViewSubjectPipeline(vo,numInputs), iParallelWorker(vo->GetViewModule()->GetParallelManager()), mId(id), mWorkPipelines(0)
{
	int j;
	mConfigureTime = 0UL;

	//
	//  Create the master thread which is always present
	//
	iViewSubjectPipeline *f = this->GetViewSubject()->CreatePipeline(mId); IERROR_ASSERT(f);
	mWorkPipelines.Add(f);

#ifdef I_CHECK1
	if(this->GetNumberOfInputs() != f->GetNumberOfInputs())
	{
		IERROR_FINISH("Misconfiguration in iViewSubjectParallelPipeline.");
	}
#endif

	vtkDataSet* tmp;
	this->SetNumberOfOutputs(f->GetNumberOfOutputs());
	for(j=0; j<f->GetNumberOfOutputs(); j++)
	{
		tmp = f->GetOutput(j)->NewInstance(); IERROR_ASSERT(tmp);
		this->SetNthOutput(j,tmp);
		tmp->Delete();
	}

	//
	//  Create data manager
	//
	mDataManager = iViewSubjectPipelineDataManager::New(vo); IERROR_ASSERT(mDataManager);
}


iViewSubjectParallelPipeline::~iViewSubjectParallelPipeline()
{
	while(mWorkPipelines.Size() > 0) mWorkPipelines.RemoveLast()->Delete();
	mDataManager->Delete();
}


void iViewSubjectParallelPipeline::Modified()
{
	int i;
	for(i=0; i<mWorkPipelines.Size(); i++) 
	{
		mWorkPipelines[i]->Modified();
	}
	this->iViewSubjectPipeline::Modified();
}


void iViewSubjectParallelPipeline::UpdateContents(int n, int info)
{
	int i;
	for(i=0; i<mWorkPipelines.Size(); i++) 
	{
		mWorkPipelines[i]->UpdateContents(n,info);
	}
	this->Modified();
}


void iViewSubjectParallelPipeline::Reconfigure()
{
	//
	//  Actual reconfiguring is done at run-time
	//
	mConfigureTime = 0UL;
}


void iViewSubjectParallelPipeline::ConfigurePipelines(vtkDataSet *input)
{
	int i, np = this->GetManager()->GetNumberOfProcessors();

	if(np < 1)
	{
		IERROR_FINISH("Cannot execute on 0 processors.");
		return;
	}

	//
	//  Delete extra pipelines
	//
	while(mWorkPipelines.Size() > np) mWorkPipelines.RemoveLast()->Delete();
	//
	//  Create missing ones
	//
	while(mWorkPipelines.Size() < np)
	{
		iViewSubjectPipeline *f = this->GetViewSubject()->CreatePipeline(mId); IERROR_ASSERT(f);
		mWorkPipelines.Add(f);
	}

	//
	//  Bring all pipelines up to date
	//
	for(i=0; i<mWorkPipelines.Size(); i++) mWorkPipelines[i]->UpdateContents(_All);

	//
	//  Assign global inputs to parallel filters
	//
 	for(i=0; i<mWorkPipelines.Size(); i++)
	{
		mWorkPipelines[i]->SetGlobalInput(input);
	}

	//
	//  Ask data manager to distribute the data 
	//
	iArray<vtkDataSet*> workInputs(0);
 	workInputs.Resize(mWorkPipelines.Size());

	this->mDataManager->DistributeData(input,workInputs);

	for(i=0; i<mWorkPipelines.Size(); i++)
	{
		mWorkPipelines[i]->SetNthInput(0,workInputs[i]);
		workInputs[i]->Delete();
	}

	//
	//  If the filters accept more than one input, just pass other inputs
	//
	if(this->GetNumberOfInputs() > 1)
	{
		int j;
		for(i=0; i<mWorkPipelines.Size(); i++)
		{
			for(j=1; j<this->GetNumberOfInputs(); j++) this->mWorkPipelines[i]->SetNthInput(j,this->GetInputs()[j]);
		}
	}

	this->iViewSubjectPipeline::Modified();
	mConfigureTime = this->GetMTime();
}


int iViewSubjectParallelPipeline::ExecuteStep(int, iParallel::ProcessorInfo &p)
{
	if(p.NumProcs != mWorkPipelines.Size()) return 1;

	this->mWorkPipelines[p.ThisProc]->Update();
	return 0;
}


void iViewSubjectParallelPipeline::InitExecution()
{
	vtkDataSet *input = this->GetInput();
	this->TransferInputInformation(input);
}


void iViewSubjectParallelPipeline::TransferInputInformation(vtkDataSet *input)
{
	input->SetUpdateExtentToWholeExtent();
}


void iViewSubjectParallelPipeline::ProduceOutput()
{
	int i, j;

	vtkDataSet *input = this->GetInput();

	for(j=0; j<this->GetNumberOfOutputs(); j++)
	{
		this->GetOutput(j)->Initialize();
	}

	this->TransferInputInformation(input);

	//
	//  Do we need to re-distribute the data?
	//
	if(mWorkPipelines.Size()!=this->GetManager()->GetNumberOfProcessors() || input->GetMTime()>this->mConfigureTime)// || this->GetMTime()>this->mConfigureTime) // the latter is needed when DataTypes are changed
	{
		this->ConfigurePipelines(input);
	}

	//
	//  Now run the pipelines
	//
	if(mWorkPipelines.Size() == 1)
	{
		//
		//  Serial execution
		//
		this->mWorkPipelines[0]->Update();
	}
	else
	{
		//
		//  Parallel execution
		//
		this->ParallelExecute(0);
//		mWorkPipelines[0]->Update();
	}

	if(this->GetAbortExecute()) return;

	//
	//  Ask DataManager to collect the data 
	//
	iArray<vtkDataSet*> workOutputs(0);
 	workOutputs.Resize(mWorkPipelines.Size());
	for(j=0; j<this->GetNumberOfOutputs(); j++)
	{
	 	for(i=0; i<mWorkPipelines.Size(); i++)
		{
			workOutputs[i] = mWorkPipelines[i]->GetOutput(j);
		}

		mDataManager->CollectData(workOutputs,this->GetOutput(j));
	}

	if(this->IsOptimizedForMemory())
	{
		this->RemoveInternalData();
	}
}


void iViewSubjectParallelPipeline::RemoveInternalDataForExtraComponents(bool keepOutputs)
{
	int i;

 	for(i=0; i<mWorkPipelines.Size(); i++)
	{
		mWorkPipelines[i]->RemoveInternalData(keepOutputs);
	}
}


float iViewSubjectParallelPipeline::GetContentsMemorySize() const
{
	int i;
	float s = 0.0f;

	for(i=0; i<mWorkPipelines.Size(); i++) 
	{
		s += mWorkPipelines[i]->GetMemorySize();
	}
	
	return s;
}

